/**
 * Copyright (c) 2000-2009 Liferay, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.ceb2b2000.connector.service.impl;

import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang.StringUtils;

import com.ceb2b2000.connector.NoSuchContentConnectionException;
import com.ceb2b2000.connector.model.ContentConnection;
import com.ceb2b2000.connector.service.base.ContentConnectionLocalServiceBaseImpl;
import com.ceb2b2000.connector.util.Assert;
import com.ceb2b2000.connector.util.ConnectorField;
import com.ceb2b2000.connector.util.Indexer;
import com.liferay.counter.service.CounterLocalServiceUtil;
import com.liferay.portal.PortalException;
import com.liferay.portal.SystemException;
import com.liferay.portal.kernel.dao.orm.DynamicQuery;
import com.liferay.portal.kernel.dao.orm.DynamicQueryFactoryUtil;
import com.liferay.portal.kernel.dao.orm.ProjectionFactoryUtil;
import com.liferay.portal.kernel.dao.orm.PropertyFactoryUtil;
import com.liferay.portal.kernel.log.Log;
import com.liferay.portal.kernel.log.LogFactoryUtil;
import com.liferay.portal.kernel.search.BooleanClauseOccur;
import com.liferay.portal.kernel.search.BooleanQuery;
import com.liferay.portal.kernel.search.BooleanQueryFactoryUtil;
import com.liferay.portal.kernel.search.Field;
import com.liferay.portal.kernel.search.Hits;
import com.liferay.portal.kernel.search.ParseException;
import com.liferay.portal.kernel.search.SearchEngineUtil;
import com.liferay.portal.kernel.search.SearchException;
import com.liferay.portal.kernel.util.GetterUtil;
import com.liferay.portal.kernel.util.Validator;
import com.liferay.portal.util.PortalUtil;
import com.liferay.portlet.expando.model.ExpandoBridge;

/**
 * <a href="ContentConnectionLocalServiceImpl.java.html"><b><i>View Source</i></b></a>
 *
 * @author Brian Wing Shun Chan
 *
 */
public class ContentConnectionLocalServiceImpl
	extends ContentConnectionLocalServiceBaseImpl {
	
	/**
	 * Remove all connections in which the element with `classPK' and `classNameId' is
	 * involved.
	 * 
	 * @param classPK - primary key of the element
	 * @param classNameId - class id of the element
	 * 
	 * @throws SystemException if an error occurs 
	 */
	public void removeConnections(long classPK, long classNameId) throws SystemException {
		Assert.positive(classPK);
		Assert.positive(classNameId);
		
		getContentConnectionPersistence().removeByPrimaryId(classPK, classNameId);
		getContentConnectionPersistence().removeBySecondaryId(classPK, classNameId);
	}
	
	/**
	 * Increment the priority of a connection.
	 *
	 * @param companyId - connection's company
	 * @param groupId - connection's group
	 * @param primaryId - the primary item's ID
	 * @param primaryClassId - the primary item's class ID
	 * @param secondaryId - the secondary item's ID
	 * @param secondaryClassId - the secondary item's class ID
	 * 
	 * @throws PortalException if an error occurs
	 * @throws SystemException if an error occurs
	 */
	public void incrementPriority(
			long companyId, long groupId, 
			long primaryId, long primaryClassId, 
			long secondaryId, long secondaryClassId)
		throws PortalException, SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(primaryId);
		Assert.positive(secondaryId);
		Assert.positive(secondaryClassId);
		
		ContentConnection connection = getContentConnectionPersistence().findByRelation(
				companyId, groupId, primaryId, primaryClassId, secondaryId, secondaryClassId);
		
		long oldPriority = connection.getPriority();
		long newPriority = oldPriority - 1;

		updatePriority(companyId, groupId, connection, oldPriority, newPriority);
	}
	
	/**
	 * Increment the priority of a connection.
	 *
	 * @param companyId - connection's company
	 * @param groupId - connection's group
	 * @param primaryId - the primary item's ID
	 * @param primaryClassId - the primary item's class ID
	 * @param secondaryId - the secondary item's ID
	 * @param secondaryClassId - the secondary item's class ID
	 * 
	 * @throws PortalException if an error occurs
	 * @throws SystemException if an error occurs
	 */
	public void decrementPriority(
			long companyId, long groupId, 
			long primaryId, long primaryClassId, 
			long secondaryId, long secondaryClassId) 
		throws PortalException, SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(primaryId);
		Assert.positive(secondaryId);
		Assert.positive(secondaryClassId);
		
		ContentConnection connection = getContentConnectionPersistence().findByRelation(
				companyId, groupId, primaryId, primaryClassId, secondaryId, secondaryClassId);
		
		long oldPriority = connection.getPriority();
		long newPriority = oldPriority + 1;
		
		updatePriority(companyId, groupId, connection, oldPriority, newPriority);
	}

	
	private void updatePriority(
			long companyId, long groupId, ContentConnection connection, 
			long oldPriority, long newPriority)
		throws SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.notNull(connection);
		Assert.positive(oldPriority);
		Assert.nonNegative(newPriority);
		
		try {
			long primaryId = connection.getPrimaryId();
			long primaryClassId = connection.getPrimaryClassId();

			ContentConnection c = getContentConnectionPersistence().findByPriority(
					companyId, groupId, primaryId, primaryClassId, newPriority);
			
			c.setPriority(oldPriority);
			updateContentConnection(c);

			connection.setPriority(newPriority);
			updateContentConnection(connection);
		} catch (NoSuchContentConnectionException e) {
			// Already has the highest/lowest priority. Ignore.
		}
	}
	
	private long getMinPriority(long companyId, long groupId, long primaryId, long primaryClassId) throws SystemException {
		DynamicQuery query = DynamicQueryFactoryUtil.forClass(ContentConnection.class);
		
		// the lower the value the *higher* the priority.
		
		query.setProjection(ProjectionFactoryUtil.max("priority"));
		
		List result = dynamicQuery(query);
		
		return result.isEmpty() ? 1 : ((Long) result.iterator().next() + 1);
	}

	public boolean existsConnection(
			long companyId, long groupId,
			long primaryId, Class<?> primaryClass,
			long secondaryId, Class<?> secondaryClass)
		throws SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(primaryId);
		Assert.notNull(primaryClass);
		Assert.positive(secondaryId);
		Assert.notNull(secondaryClass);
		
		long primaryClassId = PortalUtil.getClassNameId(primaryClass);
		long secondaryClassId = PortalUtil.getClassNameId(secondaryClass);
		
		try {
			getContentConnectionPersistence().findByRelation(
				companyId, groupId, primaryId, primaryClassId,
				secondaryId, secondaryClassId);

			return true;
		} catch (NoSuchContentConnectionException e) {
			return false;
		}
	}
	
	public void deleteContentConnection(
			long companyId, long groupId,
			long primaryId, Class<?> primaryClass,
			long secondaryId, Class<?> secondaryClass)
		throws SystemException, NoSuchContentConnectionException {

		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(primaryId);
		Assert.notNull(primaryClass);
		Assert.positive(secondaryId);
		Assert.notNull(secondaryClass);
		
		long primaryClassNameId = PortalUtil.getClassNameId(primaryClass);
		long secondaryClassNameId = PortalUtil.getClassNameId(secondaryClass);
			
		ContentConnection connection = 
			getContentConnectionPersistence().findByRelation(
					companyId, groupId, primaryId, primaryClassNameId, 
					secondaryId, secondaryClassNameId);

		long priority = connection.getPriority();
		
		deleteContentConnection(connection);

		updatePriorities(companyId, groupId, primaryId, primaryClassNameId, priority);
		
		long connectionId = connection.getConnectionId();
		
		try {
			Indexer.deleteConnection(companyId, connectionId);
		} catch (SearchException e) {
			_log.error("Error while removing index for connection " + connectionId, e);
		}
	}
	
	private void updatePriorities(long companyId, long groupId, long primaryId,
			long primaryClassNameId, long priority) throws SystemException {

		// TODO: Inneficient! this should be on one `UPDATE' query.

		DynamicQuery query = DynamicQueryFactoryUtil.forClass(ContentConnection.class);
		query.add(PropertyFactoryUtil.forName("companyId").eq(companyId));
		query.add(PropertyFactoryUtil.forName("groupId").eq(groupId));
		query.add(PropertyFactoryUtil.forName("primaryId").eq(primaryId));
		query.add(PropertyFactoryUtil.forName("primaryClassId").eq(primaryClassNameId));
		query.add(PropertyFactoryUtil.forName("priority").gt(priority));
		
		List result = dynamicQuery(query);
		
		List<ContentConnection> connections = result;
		
		for (ContentConnection cc: connections) {
			cc.setPriority(cc.getPriority() - 1);
			updateContentConnection(cc);
		}
	}

	public ContentConnection addContentConnection(
			long companyId, long groupId, long userId,
			long primaryId, Class<?> primaryClass, String primaryKeywords,
			long secondaryId, Class<?> secondaryClass, String secondaryKeywords)
		throws SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(userId);
		Assert.positive(primaryId);
		Assert.notNull(primaryClass);
		Assert.positive(secondaryId);
		Assert.notNull(secondaryClass);
		
		long primaryClassNameId = PortalUtil.getClassNameId(primaryClass);
		long secondaryClassNameId = PortalUtil.getClassNameId(secondaryClass);
		
		return addContentConnection(
				companyId, groupId, userId, 
				primaryId, primaryClassNameId, primaryKeywords, 
				secondaryId, secondaryClassNameId, secondaryKeywords);
	}
	
	public ContentConnection addContentConnection(
				Long companyId, Long groupId, Long userId,
				Long primaryId, Long primaryClassNameId, String primaryKeywords,
				Long secondaryId, Long secondaryClassNameId, String secondaryKeywords)
			throws SystemException {
		
		Assert.positive(companyId);
		Assert.positive(groupId);
		Assert.positive(userId);
		Assert.positive(primaryId);
		Assert.positive(primaryClassNameId);
		Assert.positive(secondaryId);
		Assert.positive(secondaryClassNameId);
		
		Date createDate = Calendar.getInstance().getTime();
		Date modifiedDate = createDate;
		
		long connectionId = CounterLocalServiceUtil.increment(
				ContentConnection.class.getName());
		
		ContentConnection newConnection = createContentConnection(connectionId);
		
		newConnection.setCompanyId(companyId);
		newConnection.setCreateDate(createDate);
		newConnection.setGroupId(groupId);
		newConnection.setModifiedDate(modifiedDate);
		newConnection.setPrimaryClassId(primaryClassNameId);
		newConnection.setPrimaryId(primaryId);
		
		long priority = getMinPriority(companyId, groupId, primaryId, primaryClassNameId);
		
		newConnection.setPriority(priority);
		newConnection.setSecondaryClassId(secondaryClassNameId);
		newConnection.setSecondaryId(secondaryId);
		newConnection.setUserId(userId);
		
		String truncatedPrimary = StringUtils.substring(
				primaryKeywords, 0, KEYWORD_LENGTH_LIMIT);
		
		newConnection.setPrimaryKeywords(truncatedPrimary);
		
		String truncatedSecondary = StringUtils.substring(
				secondaryKeywords, 0, KEYWORD_LENGTH_LIMIT);
		
		newConnection.setSecondaryKeywords(truncatedSecondary);
		
		addContentConnection(newConnection);
		
		try {
			Indexer.addConnection(
					companyId, groupId, userId, connectionId,
					primaryId, primaryClassNameId, primaryKeywords,
					secondaryId, secondaryClassNameId, secondaryKeywords, 
					null, newConnection.getExpandoBridge());
		} catch (SearchException e) {
			_log.error("Error while adding index for connection " + connectionId, e);
		}
		
		return newConnection;
	}
	
	public void reIndex(String className, long classPK) throws SearchException {
		try {
			ContentConnection connection = getContentConnection(classPK);
			
			reIndex(connection);
		} catch (PortalException e) {
			throw new SearchException(e);
		} catch (SystemException e) {
			throw new SearchException(e);
		}
	}

	public void reIndex(String[] ids) throws SearchException {
		try {
			if (!SearchEngineUtil.isIndexReadOnly()) {
				long companyId = GetterUtil.getLong(ids[0]);
				
				Collection<ContentConnection> cc = 
					getContentConnectionPersistence().findByCompanyId(companyId);
				
				for (ContentConnection c: cc) {
					reIndex(c);
				}
			}
		} catch (SystemException e) {
			throw new SearchException(e);
		}
	}
	
	public void reIndex(ContentConnection connection) throws SearchException {
		long companyId = connection.getCompanyId();
		long groupId = connection.getGroupId();
		long userId = connection.getUserId();
		String[] tagsEntries = {};
		long connectionId = connection.getConnectionId();
		ExpandoBridge expandoBridge = connection.getExpandoBridge();
		
		long primaryId = connection.getPrimaryId();
		long primaryClassId = connection.getPrimaryClassId();
		String primaryKeywords = connection.getPrimaryKeywords();
		long secondaryId = connection.getSecondaryId();
		long secondaryClassId = connection.getSecondaryClassId();
		String secondaryKeywords = connection.getSecondaryKeywords();
		
		Indexer.updateConnection(
				companyId, groupId, userId, connectionId,
				primaryId, primaryClassId, primaryKeywords, 
				secondaryId, secondaryClassId, secondaryKeywords,
				tagsEntries, expandoBridge);		
	}
	
	public Hits search(
			Long companyId, Long groupId, Long userId, Long ownerUserId,
			Long primaryClassNameId, String keywords, Integer start, Integer end)
		throws SystemException {
		
		return search(
				companyId, groupId, userId, ownerUserId,
				0L, primaryClassNameId, 0L, 0L, keywords, start, end);
	}

	public Hits search(
			Long companyId, Long groupId, Long userId, Long ownerUserId,
			Long primaryId, Long primaryClassNameId, Long secondaryId, Long secondaryClassNameId,
			String keywords, Integer start, Integer end)
		throws SystemException {
		
		try {
			
			BooleanQuery contextQuery = BooleanQueryFactoryUtil.create();
			
			contextQuery.addRequiredTerm(Field.PORTLET_ID, Indexer.PORTLET_ID);

			if (groupId > 0) {
				contextQuery.addRequiredTerm(Field.GROUP_ID, groupId);
			}
			
			if (ownerUserId > 0) {
				contextQuery.addRequiredTerm(Field.USER_ID, ownerUserId);
			}

			if (primaryId > 0) {
				contextQuery.addRequiredTerm(
						ConnectorField.PRIMARY_CLASS_PK, primaryId);
			}
			
			if (primaryClassNameId > 0) {
				contextQuery.addRequiredTerm(
						ConnectorField.PRIMARY_CLASS_NAME_ID, primaryClassNameId);
			}
			
			if (secondaryId > 0) {
				contextQuery.addRequiredTerm(
						ConnectorField.SECONDARY_CLASS_PK, secondaryId);
			}
			
			if (secondaryClassNameId > 0) {
				contextQuery.addRequiredTerm(
						ConnectorField.SECONDARY_CLASS_NAME_ID, secondaryClassNameId);
			}
			
			BooleanQuery searchQuery = BooleanQueryFactoryUtil.create();
			
			if (Validator.isNotNull(keywords)) {
				searchQuery.addTerm(ConnectorField.PRIMARY_KEYWORDS, keywords);
				searchQuery.addTerm(ConnectorField.SECONDARY_KEYWORDS, keywords);
				searchQuery.addTerm(Field.TAGS_ENTRIES, keywords);
			}
			
			BooleanQuery fullQuery = BooleanQueryFactoryUtil.create();
			
			fullQuery.add(contextQuery, BooleanClauseOccur.MUST);
			
			if (searchQuery.clauses().size() > 0) {
				fullQuery.add(searchQuery, BooleanClauseOccur.MUST);
			}
			
			return SearchEngineUtil.search(
					companyId, groupId, userId, ContentConnection.class.getName(),
					contextQuery, start, end);
		} catch (ParseException e) {
			throw new SystemException(e);
		} catch (SearchException e) {
			throw new SystemException(e);
		}
	}
	
	// XXX: keep in sync with model-hints.xml!!!
	private static final int KEYWORD_LENGTH_LIMIT = 500;
	
	private static final Log _log = LogFactoryUtil.getLog(
			ContentConnectionLocalServiceImpl.class);
}
